from .config import load_config_schema
from .utils import NodeModuleSpec


class TypescriptLanguageServer(NodeModuleSpec):
    node_module = key = "typescript-language-server"
    script = ["lib", "cli.js"]
    args = ["--stdio"]
    languages = [
        "javascript",
        "jsx",
        "typescript",
        "typescript-jsx",
        "typescriptreact",
        "javascriptreact",
    ]
    spec = dict(
        display_name=key,
        mime_types=[
            "application/typescript",
            "text/typescript-jsx",
            "text/javascript",
            "text/ecmascript",
            "application/javascript",
            "application/x-javascript",
            "application/ecmascript",
            "text/jsx",
        ],
        urls=dict(
            home="https://github.com/typescript-language-server/{}".format(key),
            issues="https://github.com/typescript-language-server/{}/issues".format(
                key
            ),
        ),
        install=dict(
            npm="npm install --save-dev {}".format(key),
            yarn="yarn add --dev {}".format(key),
            jlpm="jlpm add --dev {}".format(key),
        ),
        config_schema=load_config_schema(key),
    )



from ..types import LanguageServerManagerAPI
from .config import load_config_schema
from .utils import NodeModuleSpec


class BashLanguageServer(NodeModuleSpec):
    node_module = key = "bash-language-server"
    script = ["out", "cli.js"]
    fallback_script = ["bin", "main.js"]
    args = ["start"]
    languages = ["bash", "sh"]
    spec = dict(
        display_name=key,
        mime_types=["text/x-sh", "application/x-sh"],
        urls=dict(
            home="https://github.com/bash-lsp/{}".format(key),
            issues="https://github.com/bash-lsp/{}/issues".format(key),
        ),
        install=dict(
            npm="npm install --save-dev {}".format(key),
            yarn="yarn add --dev {}".format(key),
            jlpm="jlpm add --dev {}".format(key),
        ),
        config_schema=load_config_schema(key),
    )

    def solve(self, mgr: LanguageServerManagerAPI):
        new_path = mgr.find_node_module(self.node_module, *self.script)
        if new_path:
            return new_path
        return mgr.find_node_module(
            self.node_module, *self.fallback_script
        )  # pragma: no cover



from .config import load_config_schema
from .utils import NodeModuleSpec


class SQLLanguageServer(NodeModuleSpec):
    """Supports mysql, postgres and sqlite3"""

    node_module = key = "sql-language-server"
    script = ["dist", "bin", "cli.js"]
    languages = [
        "sql",
    ]
    args = ["up", "--method", "stdio"]
    spec = dict(
        display_name=key,
        mime_types=[
            "application/sql",
            "text/sql",
            "text/x-sql",
            "text/x-mysql",
            "text/x-mariadb",
            "text/x-pgsql",
        ],
        urls=dict(
            home="https://github.com/joe-re/{}".format(key),
            issues="https://github.com/joe-re/{}/issues".format(key),
        ),
        install=dict(
            npm="npm install --save-dev {}".format(key),
            yarn="yarn add --dev {}".format(key),
            jlpm="jlpm add --dev {}".format(key),
        ),
        config_schema=load_config_schema(key),
    )



import asyncio

import pytest

from ..schema import SERVERS_RESPONSE


async def assert_status_set(handler, expected_statuses, language_server=None):
    await handler.get()
    payload = handler._payload

    errors = list(SERVERS_RESPONSE.iter_errors(payload))
    assert not errors

    statuses = {
        session["status"]
        for session_server, session in payload["sessions"].items()
        if language_server is None or language_server == session_server
    }
    assert statuses == expected_statuses, payload


@pytest.mark.asyncio
async def test_start_known(known_server, handlers, jsonrpc_init_msg):
    """will a process start for a known server if a handler starts?"""
    handler, ws_handler = handlers
    manager = handler.manager

    manager.initialize()

    await assert_status_set(handler, {"not_started"})

    await ws_handler.open(known_server)
    session = manager.sessions[ws_handler.language_server]
    assert session.process is not None

    await assert_status_set(handler, {"started"}, known_server)

    await ws_handler.on_message(jsonrpc_init_msg)

    try:
        await asyncio.wait_for(
            ws_handler._messages_wrote.get(),
            120 if known_server == "julia-language-server" else 20,
        )
        ws_handler._messages_wrote.task_done()
    finally:
        ws_handler.on_close()

    assert not session.handlers
    assert not session.process

    await assert_status_set(handler, {"stopped"}, known_server)
    await assert_status_set(handler, {"stopped", "not_started"})


@pytest.mark.asyncio
async def test_start_unknown(known_unknown_server, handlers, jsonrpc_init_msg):
    """will a process not start for an unknown server if a handler starts?"""
    handler, ws_handler = handlers
    manager = handler.manager
    manager.initialize()

    await assert_status_set(handler, {"not_started"})

    await ws_handler.open(known_unknown_server)

    await assert_status_set(handler, {"not_started"})

    await ws_handler.on_message(jsonrpc_init_msg)
    await assert_status_set(handler, {"not_started"})
    ws_handler.on_close()

    assert not manager.sessions.get(ws_handler.language_server)
    await assert_status_set(handler, {"not_started"})


@pytest.mark.asyncio
async def test_ping(handlers):
    """see https://github.com/jupyter-lsp/jupyterlab-lsp/issues/458"""
    a_server = "pylsp"

    handler, ws_handler = handlers
    manager = handler.manager

    manager.initialize()

    assert ws_handler.ping_interval > 0
    # the default ping interval is 30 seconds, too long for a test
    ws_handler.settings["ws_ping_interval"] = 0.1
    assert ws_handler.ping_interval == 0.1

    assert ws_handler._ping_sent is False

    await ws_handler.open(a_server)

    assert ws_handler.ping_callback is not None and ws_handler.ping_callback.is_running
    await asyncio.sleep(ws_handler.ping_interval * 3)

    assert ws_handler._ping_sent is True

    ws_handler.on_close()



import json
import pathlib

CONFIGS = pathlib.Path(__file__).parent


def load_config_schema(key):
    """load a keyed filename"""
    return json.loads(
        (CONFIGS / "{}.schema.json".format(key)).read_text(encoding="utf-8")
    )



import os
import pathlib
import re
from urllib.parse import unquote, urlparse

RE_PATH_ANCHOR = r"^file://([^/]+|/[A-Z]:)"


def normalized_uri(root_dir):
    """Attempt to make an LSP rootUri from a ContentsManager root_dir

    Special care must be taken around windows paths: the canonical form of
    windows drives and UNC paths is lower case
    """
    root_uri = pathlib.Path(root_dir).expanduser().resolve().as_uri()
    root_uri = re.sub(
        RE_PATH_ANCHOR, lambda m: "file://{}".format(m.group(1).lower()), root_uri
    )
    return root_uri


def file_uri_to_path(file_uri):
    """Return a path string for give file:/// URI.

    Respect the different path convention on Windows.
    Based on https://stackoverflow.com/a/57463161/6646912, BSD 0
    """
    windows_path = os.name == "nt"
    file_uri_parsed = urlparse(file_uri)
    file_uri_path_unquoted = unquote(file_uri_parsed.path)
    if windows_path and file_uri_path_unquoted.startswith("/"):
        result = file_uri_path_unquoted[1:]  # pragma: no cover
    else:
        result = file_uri_path_unquoted  # pragma: no cover
    return result
